from __future__ import division
import imageio, os
from vi3o.utils import SlicedView

class ImageioVideo(object):
    def __init__(self, filename, grey=False):
        if not os.path.exists(filename):
            raise IOError("File not found: '%s'" % filename)
        if grey:
            raise NotImplementedError
        try:
            self.reader = imageio.get_reader(filename)
        except ValueError:
            self.reader = imageio.get_reader(filename, 'ffmpeg')
        self.fps = self.reader.get_meta_data()['fps']
        if self.fps == 0:
            self.fps = 25

    def __len__(self):
        return len(self.reader)

    def __getitem__(self, item):
        if isinstance(item, slice):
            return SlicedView(self, item)
        img = self.reader.get_data(item)
        img.index = item
        img.timestamp = item / self.fps
        return img

    @property
    def systimes(self):
        return [i / self.fps for i in range(len(self))]


import sys
from tempfile import NamedTemporaryFile

from py.test import raises

from vi3o.mkv import Mkv, lib, ffi
import os
import pickle

from vi3o.utils import index_file
from vi3o.compat import pathlib

mydir = os.path.dirname(__file__)
test_mkv = os.path.join(mydir, "t.mkv")
systime_mkv = os.path.join(mydir, "systime.mkv")
mjpg_codec_mkv = os.path.join(mydir, "test_mjpg_codec.mkv")

def test_iter():
    timestamps = []
    pixels = []
    video = Mkv(test_mkv)
    if os.path.exists(index_file(test_mkv)):
        os.unlink(index_file(test_mkv))
    for i, img in enumerate(video):
        assert img.index == i
        timestamps.append(img.timestamp)
        pixels.append(img[20,30,1])
    assert len(timestamps) == 16
    assert timestamps[0] == 0.0 # FIXME: Use systeime
    assert timestamps[1] == 0.04
    assert pixels[:6] == [84, 85, 85, 84, 84, 86]

    pixels = [video[i][20,30,1] for i in range(6)]
    assert pixels == [84, 85, 85, 84, 84, 86]

    assert video[2].timestamp == 0.08
    assert video[1].timestamp == 0.04
    assert video[2].index == 2
    assert video[0][20,30,1] == 84

    video2 = Mkv(test_mkv)
    assert video[3][20,30,1] == 84

    assert len(video)== 16
    with raises(IndexError):
        video[100]

def test_idx():
    if os.path.exists(index_file(test_mkv)):
        os.unlink(index_file(test_mkv))
    test_iter()
    test_iter()

def test_no_file():
    with raises(IOError):
        Mkv(test_mkv + 'not_there')

def test_mkv_pathlib_open():
    _ = Mkv(pathlib.Path(test_mkv))

def test_slice():
    video = Mkv(test_mkv)
    sub = video[2:5]
    assert [img[20,30,1] for img in sub] == [85, 84, 84]
    assert sub[1].timestamp == video[3].timestamp
    assert len(sub) == 3

def test_systime():
    video = Mkv(systime_mkv)
    assert video[7].systime == 1448984844.6525
    t = [img.systime for img in video]
    assert t == video.systimes

    cut = video[10:20]
    assert t[10:20] == cut.systimes

def test_getitem():
    video = Mkv(systime_mkv)
    idx = 31
    for i in range(3):
        assert video[idx-1].index == idx-1
    for i in range(3):
        assert video[idx].index == idx
    for i in range(3):
        assert video[idx+1].index == idx+1

def test_bad_file():
    fn = test_mkv + '_bad'
    if sys.version_info > (3,):
        fn = bytes(fn, "utf8")
    for bad in ["\000", "\001"]:
        with open(fn, "w") as fd:
            fd.write(bad)
        m = lib.mkv_open(fn)
        frm = ffi.new('struct mkv_frame *')
        assert lib.mkv_next(m, frm) == 0

def test_pickle():
    video = Mkv(systime_mkv)
    t8 = video[8].systime
    assert t8 > 10000
    sub = video[6::2]
    assert sub[1].systime == t8

    assert video.myiter is not None
    with NamedTemporaryFile() as tmp:
        with open(tmp.name, "wb") as fd:
            pickle.dump(sub, fd)
        loaded = pickle.load(open(tmp.name, "rb"))

    assert loaded[1].systime == t8

def test_mjpg_codec():
    systimes = []
    video = Mkv(mjpg_codec_mkv)
    for img in video:
        systimes.append(img.systime)
        assert img.shape == (300, 480, 3)
    assert video[0].systime == 1539001990.82
    assert video[2].systime == 1539001992.82
    assert video[1].systime == 1539001991.82
    assert systimes == [1539001990.82, 1539001991.82, 1539001992.82]

def test_mjpg_codec_grey():
    systimes = []
    video = Mkv(mjpg_codec_mkv, grey=True)
    for img in video:
        systimes.append(img.systime)
        assert img.shape == (300, 480)
    assert systimes == [1539001990.82, 1539001991.82, 1539001992.82]


from cffi import FFI
import os

mydir = os.path.abspath(os.path.dirname(__file__))

ffi = FFI()
ffi.cdef("""
        struct mkv {
            int width, height;
            char mac[13];
            int64_t systime_offset_sum, systime_offset_count, systime_offset;
            uint8_t *codec_private;
            size_t codec_private_len;
            char *codec_id;
            ...;
        };
        struct mkv_frame {
            uint64_t pts, systime;
            uint8_t *data;
            unsigned long len;
            unsigned long offset;
            int key_frame;
        };

        struct mkv *mkv_open(char *filename);
        void mkv_close(struct mkv *s);
        int mkv_next(struct mkv *s, struct mkv_frame *frm);
        void mkv_seek(struct mkv *s, unsigned long offset);

        struct decode;
        struct decode *decode_open(struct mkv *m);
        void decode_close(struct decode *p);
        int decode_frame(struct decode *p, struct mkv_frame *frm, uint8_t *img, uint64_t *ts, int grey);
        int64_t mkv_estimate_systime_offset(struct mkv *s);

         """)
ffi.set_source("vi3o._mkv", '#include "src/decode.h"',
               include_dirs=[mydir],
               sources=["src/mkv.c", "src/decode.c"],
               # cflags=["-g"],
               libraries=["avcodec", "swscale"])

if __name__ == '__main__':
    ffi.compile()



from cffi import FFI
import os

mydir = os.path.abspath(os.path.dirname(__file__))

ffi = FFI()
ffi.cdef("""
        struct mjpg {
            int width, height;
            unsigned char *pixels;
            unsigned int timestamp_sec;
            unsigned int timestamp_usec;
            long start_position_in_file, stop_position_in_file;
            char hwid[32];
            char serial[32];
            char firmware[32];
            ...;
        };
        enum {OK, ERROR_FILENOTFOUND, ERROR_ILLEGALARGUMENT, ERROR_FILEFORMAT,
              ERROR_EOF, ERROR_FAIL};
        enum {IMORDER_PLANAR, IMORDER_PLANAR_SUBX, IMORDER_PLANAR_SUBXY,
              IMORDER_INTERLEAVED};
        enum {IMTYPE_GRAY, IMTYPE_YCbCr, IMTYPE_RGB, IMTYPE_BGR};

        int mjpg_open(struct mjpg *m, char *name, int type, int dataOrder);
        int mjpg_open_buffer(struct mjpg *m, uint8_t *buf, int len, int type, int dataOrder);
        int mjpg_next_head(struct mjpg *m);
        int mjpg_next_data(struct mjpg *m);
        int mjpg_close(struct mjpg *m);
        int mjpg_seek (struct mjpg *m, long offset);

         """)
ffi.set_source("vi3o._mjpg", '#include "src/mjpg.h"',
               include_dirs=[mydir],
               sources=["src/mjpg.c"],
               libraries=["jpeg"], )

if __name__ == '__main__':
    ffi.compile()




import sys, os, hashlib
import numpy as np
import vi3o
from vi3o import compat

if sys.version_info > (3,):
    xrange = range

class Frame(np.ndarray):
    pass

class SlicedView(object):
    def __init__(self, parent, indexes, properties=()):
        self.parent = parent
        self.range = xrange(*indexes.indices(len(parent)))
        self.properties = properties

    def __getitem__(self, item):
        return self.parent[self.range[item]]

    def __len__(self):
        return len(self.range)

    def __getattr__(self, item):
        if item in self.properties:
            return self.properties[item](self.range)
        return getattr(self.parent, item)

    def __getstate__(self):
        state = dict(self.__dict__)
        del state['properties'] # FIXME
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.properties = ()

class VideoFilter:
    """Apply a filter on the output frames from a video

    VideoFilter objects can be used as regular vi3o videos but applies a filter to
    the Frame object when iterating or accessing it.

    E.g, create a video that only is using the top half of the frames:

        .. code-block:: python

            video = VideoFilter(fname, lambda frame: frame[:frame.shape[0]//2])

    E.g. offset the timestamp of each frame by one hour:

        .. code-block:: python

            def offset_time(frame):
                frame.systime += 3600
                return frame

            video = VideoFilter(fname, offset_time)
    """

    def __init__(self, video, filter_function, **kwargs):
        if isinstance(video, compat.basestring):
            video = vi3o.Video(video, **kwargs)
        self.video = video
        self.filter_function = filter_function

    def _sliced_systimes(self, range):
        """Systimes property might not be implemented, depends on base video class
        """
        return [self.systimes[i] for i in range]

    def __iter__(self):
        for f in self.video:
            yield self._filter(f)

    def __len__(self):
        return len(self.video)

    def __getitem__(self, item):
        if isinstance(item, slice):
            try:
                return SlicedView(self, item, {"systimes": self._sliced_systimes})
            except AttributeError:
                return SlicedView(self, item)
        return self._filter(self.video[item])

    def __getattr__(self, name):
        return getattr(self.video, name)

    def _filter(self, org_frame):
        img = self.filter_function(org_frame)
        frame = copy_missing_metadata(img, org_frame)
        return frame

def copy_missing_metadata(dst_img, template_frame):
    """Copy missing meta data from template Frame object
    """
    def try_set_attribute(frame, attribute):
        # Try first to keep from the input image
        if hasattr(dst_img, attribute):
            setattr(frame, attribute, getattr(dst_img, attribute))
            return
        # Try to copy from the template frame
        # properties for meta data might differ with video type...
        if hasattr(template_frame, attribute):
            setattr(frame, attribute, getattr(template_frame, attribute))
            
    dst_frame = dst_img.view(Frame)
    for attribute in ("timestamp", "systime", "pts", "index"):
        try_set_attribute(dst_frame, attribute)
    return dst_frame

cache_dir = os.path.join(os.path.expanduser('~'), ".cache", "vi3o")

def index_file(fn, extradata=None):
    stats = os.stat(fn)
    key = str((os.path.abspath(fn), stats.st_size, stats.st_mtime, extradata))
    key = hashlib.md5(key.encode()).hexdigest()
    path = os.path.join(cache_dir, key + '.idx')
    d = os.path.dirname(path)
    try:
        os.makedirs(d)
    except OSError:
        pass
    return path



"""
:mod:`vi3o.mjpg` --- Motion JPEG video loading
==============================================
"""

import json
import os, sys
from vi3o.utils import SlicedView, index_file, Frame
from vi3o._mjpg import ffi, lib

class Mjpg(object):
    """
    If a filename that ends with .mjpg is passed to :func:`vi3o.Video` this kind of object
    is returned. It has a few additional format specific properties:
    """
    def __init__(self, filename, grey=False):
        # Be compatible with pathlib.Path filenames
        filename = str(filename).encode('utf-8')
        self.filename = filename
        self.grey = grey
        open(filename).close()
        self._myiter = None
        self._index = None

    def __iter__(self):
        return MjpgIter(self.filename, self.grey)

    @property
    def myiter(self):
        if self._myiter is None:
            self._myiter = iter(self)
        return self._myiter

    @property
    def offset(self):
        if self._index is None:
            idx = index_file(self.filename, self.grey)
            if os.path.exists(idx):
                self._index = json.load(open(idx))
            else:
                self._index = [self.myiter.m.start_position_in_file for img in self.myiter]
                with open(idx, 'w') as fd:
                    json.dump(self._index, fd)
        return self._index

    @property
    def systimes(self):
        raise NotImplementedError

    def _sliced_systimes(self, range):
        return [self.systimes[i] for i in range]

    def __getitem__(self, item):
        if isinstance(item, slice):
            return SlicedView(self, item, {'systimes': self._sliced_systimes})
        if (item < 0):
            item += len(self)
        lib.mjpg_seek(self.myiter.m, self.offset[item])
        self.myiter.fcnt = item
        return self.myiter.next()

    def __len__(self):
        return len(self.offset)

    @property
    def hwid(self):
        """
        The Axis hardware id of the camera that made this recording.
        """
        self.myiter.next()
        return ffi.string(self.myiter.m.hwid)

    @property
    def serial_number(self):
        """
        The Axis serial number or mac address of the camera that made this recording.
        """
        self.myiter.next()
        return ffi.string(self.myiter.m.serial)

    @property
    def firmware_version(self):
        """
        The firmware version running in the camera when it made this recording.
        """
        self.myiter.next()
        return ffi.string(self.myiter.m.firmware)


class MjpgIter(object):
    def __init__(self, filename, grey=False):
        self.m = ffi.new("struct mjpg *")
        self.fcnt = 0
        if grey:
            r = lib.mjpg_open(self.m, filename, lib.IMTYPE_GRAY, lib.IMORDER_INTERLEAVED)
            self.channels = 1
        else:
            r = lib.mjpg_open(self.m, filename, lib.IMTYPE_RGB, lib.IMORDER_INTERLEAVED)
            self.channels = 3
        if r != lib.OK:
            raise IOError("Failed to open: " + filename)

    def __iter__(self):
        return self

    def next(self):
        r = lib.mjpg_next_head(self.m)
        if r != lib.OK:
            raise StopIteration
        if self.channels == 1:
            shape = (self.m.height, self.m.width)
        else:
            shape = (self.m.height, self.m.width, self.channels)
        img = Frame(shape, 'B')
        assert img.__array_interface__['strides'] is None
        self.m.pixels = ffi.cast('unsigned char *', img.__array_interface__['data'][0])

        r = lib.mjpg_next_data(self.m)
        if r != lib.OK:
            raise StopIteration

        # img = img.reshape(shape).view(type=Frame)
        img.timestamp = self.m.timestamp_sec + self.m.timestamp_usec / 1000000.0
        img.systime = img.timestamp
        img.index = self.fcnt
        self.fcnt += 1
        return img

    def __next__(self):
        return self.next()

    def __del__(self):
        lib.mjpg_close(self.m)

def jpg_info(filename):
    """
    Reads a single jpeg image from the file *filename* and extracts the Axis user data header.
    The information it contains is returned as a dict with the keys "hwid", "serial_numer" and
    "firmware_version".
    """
    if sys.version_info > (3,):
        filename = bytes(filename, "utf8")
    m = ffi.new("struct mjpg *")
    r = lib.mjpg_open(m, filename, lib.IMTYPE_GRAY, lib.IMORDER_PLANAR)
    lib.mjpg_next_head(m)
    res = {'hwid': ffi.string(m.hwid),
           'serial_number': ffi.string(m.serial),
           'firmware_version': ffi.string(m.firmware),
           'timestamp': m.timestamp_sec + m.timestamp_usec / 1000000.0}
    lib.mjpg_close(m)
    return res






"""
:mod:`vi3o.opencv` --- Fallbacks on OpenCV functionality
========================================================
"""

import cv2
from vi3o.image import ptpscale
from vi3o.utils import Frame, SlicedView
import os
import numpy as np

for attr in ('CAP_PROP_FRAME_COUNT', 'CAP_PROP_POS_FRAMES', 'CAP_PROP_POS_MSEC'):
    if not hasattr(cv2, attr):
        setattr(cv2, attr, getattr(cv2.cv, 'CV_' + attr))

class CvVideo(object):
    def __init__(self, filename, grey=False):
        if not os.path.exists(filename):
            raise IOError("File not found: '%s'" % filename)
        self.grey = grey
        self.capture = cv2.VideoCapture(filename)

    def __len__(self):
        return int(self.capture.get(cv2.CAP_PROP_FRAME_COUNT))

    def __next__(self):
        return self.next()

    def next(self):
        index = int(self.capture.get(cv2.CAP_PROP_POS_FRAMES))
        timestamp = self.capture.get(cv2.CAP_PROP_POS_MSEC) / 1000.0
        ret, frame = self.capture.read()
        if not ret:
            raise StopIteration
        if self.grey:
            img = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        else:
            img = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        img = img.view(Frame)
        img.index = index
        img.timestamp = timestamp
        return img


    def __iter__(self):
        return self

    def __getitem__(self, item):
        if isinstance(item, slice):
            return SlicedView(self, item)
        self.capture.set(cv2.CAP_PROP_POS_FRAMES, item)
        return self.next()

    def __del__(self):
        self.capture.release()

class CvOut(object):
    """
     Video file output using the same interface as the DebugViewer, but encodes the output to a video file
     such as .avi or .mkv instead. The *filname* specified with be overwritten and the frame rate can be
     specified using the *fps* argument. The video will be encoded using H264 if available and fall back on
     DIVX if not. To ensure that the file is proberly closed, this object can be used as a context manager:

     .. code-block:: python

        from vi3o.opencv import CvOut

        with CvOut("result.avi") as avi:
            avi.view(frame1)
            avi.view(frame2)
            ...

    """
    def __init__(self, filename, fps=25):
        self.filename = filename
        self.video = None
        self.fps = fps

    def view(self, img, scale=False):
        if img.dtype == 'bool':
            img = img.astype('B')
        if scale:
            img = ptpscale(img)
        if img.dtype != 'B':
            img = np.minimum(np.maximum(img, 0), 255).astype('B')

        if len(img.shape) == 2:
            img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
        else:
            img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)

        if self.video is None:
            height, width, _ = img.shape
            for codec in [875967048, 1482049860, -1]: #[cv2.cv.FOURCC(*"H264"), cv2.cv.FOURCC(*"DIVX"), -1]:
                self.video = cv2.VideoWriter(self.filename, codec, self.fps, (width, height))
                if self.video.isOpened():
                    break
        self.video.write(img)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__del__()

    def __del__(self):
        if self.video:
            self.video.release()
        self.video = None

if __name__ == '__main__':
    import numpy as np
    avi = CvOut("/tmp/t.avi")
    for i in range(100):
        avi.view(np.zeros((480, 640, 3)))
    del avi



"""
:mod:`vi3o.netcam` --- Live camera handling
===========================================
"""

import base64
import sys
import os

if sys.version_info > (3,):
    from io import BytesIO as StringIO
else:
    from cStringIO import StringIO

import requests
from requests.auth import HTTPBasicAuth, HTTPDigestAuth

from vi3o.image import imread
from vi3o.utils import Frame


class AxisCam(object):
    """
    Loads an mjpg stream directly from an Axis camera with hostname *ip* with the
    resolution *width*x*height* using the *username* and *password* as credntials.
    If *no_proxy* is True, the proxy settings from the environment will be ignored
    and any other keyword parameter will be passed on to the camera as a VAPIX
    parameter.
    """
    def __init__(self, ip, width=None, height=None, username=None, password=None, no_proxy=False, **kwargs):


        if no_proxy:
            os.environ['NO_PROXY'] = ip
            os.environ['no_proxy'] = ip
        mjpg_url = 'http://' + ip + '/mjpg/video.mjpg'
        if width is not None:
            mjpg_url += '?resolution=%dx%d' % (width, height)
        for k, v in kwargs.items():
            mjpg_url += '&%s=%s' % (k, v)

        r = requests.get(mjpg_url, auth=HTTPDigestAuth(username, password), stream=True)
        r.raise_for_status()

        self._fd = r.raw
        self.fcnt = 0

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        headers = {}
        while True:
            l = self._fd.readline()
            if not l:
                raise StopIteration
            if not l.strip() and headers:
                break
            if b':' in l:
                i = l.index(b':')
                headers[l[:i]] = l[i+1:].strip()

        data = self._fd.read(int(headers[b'Content-Length']))
        img = imread(StringIO(data)).view(Frame)
        img.index = self.fcnt
        self.fcnt += 1
        img.timestamp = img.systime = -1 # FIXME
        return img

if __name__ == '__main__':
    from vi3o import view
    for img in AxisCam("192.168.0.90", username="root", password="pass", no_proxy=True):
        view(img)


from py.test import raises
import vi3o
from vi3o.mjpg import Mjpg
from vi3o.mkv import Mkv
from vi3o.utils import VideoFilter, Frame
from test.util import mock
import numpy as np
import os

mydir = os.path.dirname(__file__)
test_mjpg = os.path.join(mydir, "t.mjpg")
systime_mkv = os.path.join(mydir, "systime.mkv")

def test_slice_not_implemented_systime():
    """Systimes property not implemented for all videos

    propagate this correctly, but do allow slices..
    """
    class Dummy(mock.MagicMock):
        @property
        def systimes(self):
            raise NotImplementedError()
    video_mock = Dummy(spec_set=Mjpg)
    video_mock.__len__.return_value = 10
    video = VideoFilter(video_mock, lambda f: f[:])
    with raises(NotImplementedError):
        _ = video.systimes
    cut = video[2:4]
    assert len(cut) == 2
    with raises(NotImplementedError):
        _ = cut.systimes

def test_slice_implemented_systime():
    """Systimes property should follow a slice
    """
    video_mock = mock.MagicMock(spec_set=Mjpg)
    video_mock.__len__.return_value = 4
    video_mock.systimes = [1, 2, 4, 10]
    video = VideoFilter(video_mock, lambda f: f[:])
    assert video.systimes == [1, 2, 4, 10]
    cut = video[1:3]
    assert cut.systimes == [2, 4]

def test_iter():
    timestamps = []
    pixels = []
    video = VideoFilter(Mjpg(test_mjpg), lambda f: f[:50,:50])
    for i, img in enumerate(video):
        assert img.index == i
        timestamps.append(img.timestamp)
        assert img.timestamp == img.systime
        pixels.append(img[20,30,1])
    assert len(timestamps) == 16
    assert timestamps[1] == 1445859308.97
    assert pixels[:7] == [84, 83, 84, 86, 86, 84, 84]

    assert video[1].timestamp == 1445859308.97
    assert video[2].index == 2
    assert video[0][20,30,1] == 84
    assert len(video)== 16
    with raises(IndexError):
        video[100]

@mock.patch.object(vi3o, 'Video')
def test_no_file(video_mock):
    def raise_error(filename, **kwargs):
        raise IOError("Mock error for %s" % filename)

    video_mock.side_effect = raise_error
    with raises(IOError):
        VideoFilter("dummy", lambda f: f)

def test_slice():
    video = VideoFilter(Mjpg(test_mjpg), lambda f: f)
    sub = video[2:5]
    assert [img[20,30,1] for img in sub] == [84, 86, 86]
    assert sub[1].timestamp == video[3].timestamp
    assert len(sub) == 3

def test_grey():
    video = VideoFilter(Mjpg(test_mjpg, grey=True), lambda f: f)
    for img in video:
        assert img.shape == (120, 160)

    video = VideoFilter(Mjpg(test_mjpg, grey=True), lambda f: f[:f.shape[0]//2])
    for img in video:
        assert img.shape == (60, 160)

def test_open_by_fname():
    video = VideoFilter(str(test_mjpg), lambda f: f)
    assert video[0].shape == (120, 160, 3)

def test_change_metadata():
    def offset_time(frame):
        frame.systime += 3600
        return frame

    pixels = []
    video1 = Mjpg(test_mjpg)
    video2 = VideoFilter(video1, offset_time)
    for img, img_filtered in zip(video1, video2):
        assert img.systime == img_filtered.systime - 3600
        assert img.timestamp == img_filtered.timestamp
        assert np.all(img[:7] == img_filtered[:7])

def test_change_metadata_on_copy():
    def offset_time(org_frame):
        frame = org_frame[:]
        frame.systime = org_frame.systime + 3600
        return frame

    pixels = []
    video1 = Mjpg(test_mjpg)
    video2 = VideoFilter(video1, offset_time)
    for img, img_filtered in zip(video1, video2):
        assert img.systime == img_filtered.systime - 3600
        assert img.timestamp == img_filtered.timestamp
        assert np.all(img[:7] == img_filtered[:7])

def test_mkv_systime():
    video = VideoFilter(Mkv(systime_mkv), lambda f: f[:])
    assert video[7].systime == 1448984844.6525
    t = [img.systime for img in video]
    assert t == video.systimes

    cut = video[10:20]
    assert t[10:20] == cut.systimes

def test_mkv_getitem():
    video = VideoFilter(Mkv(systime_mkv), lambda f: f[:])
    idx = 31
    for i in range(3):
        assert video[idx-1].index == idx-1
    for i in range(3):
        assert video[idx].index == idx
    for i in range(3):
        assert video[idx+1].index == idx+1

def _get_np_pointer(arr):
    return arr.__array_interface__["data"][0]

@mock.patch.object(vi3o, "Video")
def test_no_copy_filter(mocked_video):
    def filter_func(img):
        return img

    frm = Mjpg(test_mjpg)[0]
    mocked_video.return_value = [frm]
    
    video = VideoFilter("dummy", filter_func)
    assert _get_np_pointer(video[0]) == _get_np_pointer(frm)

@mock.patch.object(vi3o, "Video")
def test_copy_filter(mocked_video):
    def filter_func(img):
        return img.copy()

    frm = Mjpg(test_mjpg)[0]
    mocked_video.return_value = [frm]

    video = VideoFilter("dummy", filter_func)
    assert _get_np_pointer(video[0]) != _get_np_pointer(frm)
    assert video[0].timestamp == frm.timestamp



from threading import RLock

import pyglet
from pyglet.window import key as keysym
from pyglet.gl import glTexParameteri, GL_TEXTURE_MAG_FILTER, GL_TEXTURE_MIN_FILTER, GL_NEAREST
from vi3o.image import ptpscale
import numpy as np

global_pyglet_lock = RLock()

_PYGLET_VERSION = tuple(int(v) for v in pyglet.version.split('.'))

# Compatibility fix for breaking pyglet changes in 1.3 -> 1.4
if _PYGLET_VERSION >= (1, 4, 0):
    def get_texture(img):
        return img.get_texture()
else:
    def get_texture(img):
        return img.texture


class DebugViewer(object):
    paused = False
    step_counter = 0
    zoom = 1.0
    mouse_x = mouse_y = 0
    scroll = [0, 0]
    named_viewers =  {}

    def __init__(self, name='Video'):
        with global_pyglet_lock:
            self.window = pyglet.window.Window(resizable=True, caption=name)
            self.name = name
            self.fcnt = 0
            for name in dir(self):
                if name[:3] == 'on_':
                    self.window.event(getattr(self, name))
            self.mystep = DebugViewer.step_counter
            self.fullscreen = False
            self.image = None
            self.label = pyglet.text.Label('Hello, world',
                              font_name='Times New Roman',
                              font_size=16,
                              x=10, y=0,
                              anchor_x='left', anchor_y='bottom')
            self.label.set_style('background_color', (0,0,0,255))
            self.autoflipp = True
            self.original_image_array = None
            self.force_scale = False
            self.prev_image_shape = None
            self.xcoordinate_modulus = 1

    def _dispatch_events(self):
        for window in pyglet.app.windows:
            window.switch_to()
            window.dispatch_events()
            window.dispatch_event('on_draw')

    def _dispatch_event(self, *e):
        for window in pyglet.app.windows:
            window.dispatch_event(*e)

    def _inc_fcnt(self):
        self.fcnt += 1

    def view(self, img, scale=False, intensity=None, pause=None):
        with global_pyglet_lock:
            if img.dtype == 'bool':
                img = img.astype('B')
            if intensity is None:
                intensity = img
            if scale:
                img = ptpscale(img)
            if img.dtype != 'B':
                img = np.minimum(np.maximum(img, 0), 255).astype('B')

            if pause is not None:
                DebugViewer.paused = pause

            if self.autoflipp:
                self.image_array = [(img, intensity)]
                self._inc_fcnt()
                self._view_image_array()
            else:
                self.image_array.append((img, intensity))

    def flipp(self, pause=None):
        with global_pyglet_lock:
            if pause is not None:
                DebugViewer.paused = pause
            if not self.autoflipp:
                self._inc_fcnt()
                self._view_image_array()
            else:
                self.image = None
            self.autoflipp = False
            self.image_array = []


    def _pad_height(self, img, h):
        extra = h - img.shape[0]
        assert extra >= 0
        if extra == 0:
            return img
        top = extra // 2
        shape = list(img.shape)
        shape[0]  += extra
        res = np.ones(shape, img.dtype) * 25
        res[top:top+img.shape[0]] = img
        return res

    def _stack(self, image_array):
        if max(len(img.shape) for img in image_array) == 3:
            images = [img if len(img.shape) == 3 else np.stack([img, img, img], 2)
                      for img in image_array]
        else:
            images = image_array
        h = max(img.shape[0] for img in image_array)
        images = [self._pad_height(img, h) for img in images]
        return np.hstack(images)

    def _view_image_array(self):

        if self.force_scale:
            self.original_image_array = self.image_array[:]
            for i in range(len(self.image_array)):
                img, intensity = self.image_array[i]
                img = ptpscale(intensity)
                img = np.minimum(np.maximum(img, 0), 255).astype('B')
                self.image_array[i] = (img, intensity)
        else:
            self.original_image_array = None

        self.window.set_caption(self.name + ' - %d' % self.fcnt)


        img = self._stack([img for img, _ in self.image_array])
        intensity = self._stack([ii for _, ii in self.image_array])
        self.color_mask = np.hstack([[len(ii.shape)==3] * ii.shape[1]
                                     for _, ii in self.image_array])
        # FIXME: split image_array

        if self.image is None:
            resize = True
        else:
            resize = img.shape != self.prev_image_shape
        self.prev_image_shape = img.shape

        if len(img.shape) == 3:
            f = 'RGB'
            pitch = -img.shape[1] * 3
        else:
            f = 'I'
            pitch = -img.shape[1]
        self.image = pyglet.image.ImageData(img.shape[1], img.shape[0], f, img.tostring(), pitch)

        texture = get_texture(self.image)
        glTexParameteri(texture.target,
            GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameteri(texture.target,
            GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        if resize:
            self.on_resize(self.window.width, self.window.height)

        self.intensity = intensity
        while True:
            self._dispatch_events()
            if not DebugViewer.paused:
                self.mystep = DebugViewer.step_counter
                break
            elif self.mystep < DebugViewer.step_counter:
                self.mystep += 1
                break

    def on_draw(self):
        pyglet.gl.glClearColor(0.1, 0.1, 0.1, 1.0)
        self.window.clear()
        if self.image:
            self.image.blit(self.offset[0] + self.scroll[0], self.offset[1] + self.scroll[1], 0,
                            self.scaled_size[0], self.scaled_size[1])
            x, y = DebugViewer.mouse_x, DebugViewer.mouse_y
            if 0 <= x < self.intensity.shape[1] and 0 <= y < self.intensity.shape[0]:
                ii = self.intensity[int(y), int(x)]
                if not self.color_mask[int(x)] and len(ii.shape) > 0:
                    ii = ii[0]
                self.label.text = 'x: %4d   y: %4d   I: %s' % (x, y, ii)
            else:
                self.label.text = ''
        self.label.draw()

        self.window.flip()

    # Keys not repeated when held down
    def on_key_press(self, key, modifiers):
        if key == keysym.Q or key == keysym.ESCAPE:
            exit(0)
        elif key == keysym.F:
            self.fullscreen = not self.fullscreen
            self.window.set_fullscreen(self.fullscreen)
        elif key == keysym.Z:
            DebugViewer.zoom = 1.0
            DebugViewer.scroll = [0, 0]
            self._dispatch_event('on_resize', self.window.width, self.window.height)
        elif key == keysym.S:
            if self.force_scale:
                assert self.original_image_array is not None
                self.image_array = self.original_image_array
                self.force_scale = False
            else:
                self.force_scale = True
            self._view_image_array()
        elif key == keysym.D:
            import pdb; pdb.set_trace()

    # Keys repeated when held down
    def on_text(self, char):
        if char == " ":
            DebugViewer.paused = not DebugViewer.paused
        elif char in "\r\n":
            DebugViewer.step_counter += 1
        elif char in '123456789':
            self.xcoordinate_modulus = int(char)


    def on_mouse_motion(self, x, y, dx=None, dy=None):
        x = int((x - self.offset[0] - self.scroll[0]) / self.scale)
        if self.image is not None:
            x = x % (self.image.width / self.xcoordinate_modulus)
        y = self.image.height - int((y - self.offset[1] - self.scroll[1]) / self.scale) - 1
        DebugViewer.mouse_x, DebugViewer.mouse_y = x, y

    def on_resize(self, winw, winh):
        if self.image:
            lh = self.label.content_height
            self.scale = sc = min(float(winw) / self.image.width,
                                  float(winh - lh) / self.image.height) * self.zoom
            w, h = int(sc * self.image.width), int(sc * self.image.height)
            self.offset = (self.window.width - w) / 2, (self.window.height - lh - h) / 2 + lh
            self.scaled_size = w, h

    def on_mouse_scroll(self, x, y, scroll_x, scroll_y):
        self.on_mouse_motion(x, y)
        prex, prey = DebugViewer.mouse_x, DebugViewer.mouse_y
        DebugViewer.zoom *= 1.25 ** scroll_y
        self._dispatch_event('on_resize', self.window.width, self.window.height)
        self.on_mouse_motion(x, y)
        self.scroll[0] += (DebugViewer.mouse_x - prex) * self.scale
        self.scroll[1] -= (DebugViewer.mouse_y - prey) * self.scale
        self.on_mouse_motion(x, y)
        self.clicking = False

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        DebugViewer.scroll[0] += dx
        DebugViewer.scroll[1] += dy
        self.clicking = False

    def on_mouse_press(self, x, y, button, modifiers):
        self.clicking = True

    def on_mouse_release(self, x, y, button, modifiers):
        if self.clicking:
            print(DebugViewer.mouse_x, DebugViewer.mouse_y)

if __name__ == '__main__':
    from vi3o import Video, view
    import sys

    if len(sys.argv) < 2:
        print("Usage: python -mvi3o.debugview <video file>")
        exit(-1)

    for fn in sys.argv[1:]:
        if fn.split('.')[-1].lower() in ('mkv', 'mjpg', 'avi', 'mov', 'mp4'):
            for img in Video(sys.argv[1]):
                view(img)
        else:
            from vi3o.image import imread, imshow
            imshow(imread(fn))



import os
from vi3o import SyncedVideos

mydir = os.path.dirname(__file__)
test_mkvs = [os.path.join(mydir, f) for f in ['a.mkv', 'b.mkv', 'c.mkv']]


def test_sync():
    count = 0
    videos = SyncedVideos(*test_mkvs)
    systimes = []
    for a, b, c in videos:
        assert abs(a.systime - b.systime) < 0.06
        assert abs(b.systime - c.systime) < 0.06
        assert abs(a.systime - c.systime) < 0.06
        count += 1
        systimes.append((a.systime, b.systime, c.systime))
    assert count == 104
    assert videos.systimes == systimes

    assert len(videos) == 104
    assert len(videos[:10]) == 10
    assert len(videos[-10:]) == 10

    for i in range(104):
        a, b, c = videos[i]
        assert systimes[i] == (a.systime, b.systime, c.systime)

    skip50 = videos[50:]
    for i in range(104-50):
        a, b, c = skip50[i]
        assert systimes[i+50] == (a.systime, b.systime, c.systime)


    last10 = videos[-10:]
    for i in range(10):
        a, b, c = last10[i]
        assert systimes[i + 94] == (a.systime, b.systime, c.systime)

    assert last10.videos is videos.videos

    assert last10.systimes == systimes[-10:]
    assert last10.indexes == videos.indexes[-10:]




"""
    Interprets the storage format for recordings in AXIS video
    storage format,  e.g. video stored on a NAS or SD card.
 """
from __future__ import unicode_literals, print_function
import collections
import datetime
import sys
import xml.etree.ElementTree

import vi3o
from vi3o import mkv
from vi3o import cat
from vi3o import compat

if sys.version_info >= (3, 5, 0):
    from typing import Any, List, Union, Dict

# Representation of data in XML block of a recording
RecordingBlock = collections.namedtuple(
    "RecordingBlock", ["start", "stop", "filename", "status"]
)
# Representation of data in XML of recording
RecordingMetadata = collections.namedtuple(
    "RecordingMetadata",
    [
        "recording_id",
        "channel",
        "start",
        "stop",
        "blocks",
        "width",
        "height",
        "framerate",
    ],
)


def _parse_axis_xml_timestamp(string):
    # type: (str) -> float
    return compat.utc_datetime_to_epoch(
        datetime.datetime.strptime(string, "%Y-%m-%dT%H:%M:%S.%fZ")
    )


def _read_xml(path):
    # type: (pathlib.Path) -> Any
    tree = xml.etree.ElementTree.parse(str(path))
    return tree.getroot()


def _read_recording_block(path):
    # type: (pathlib.Path) -> RecordingBlock
    root = _read_xml(path)

    filename = path.parent / (path.stem + ".mkv")
    if not filename.exists():
        filename = path.parent / (path.stem + ".mjpg")

    if not filename.exists():
        raise compat.FileNotFoundError(
            "Could not find a video file for {}".format(path)
        )

    status = root.findall(".//Status")[0].text
    if status != "Complete":
        raise RuntimeError(
            '{}: Expected status "Complete", was: "{}"'.format(path, status)
        )

    return RecordingBlock(
        start=_parse_axis_xml_timestamp(root.findall(".//StartTime")[0].text),
        stop=_parse_axis_xml_timestamp(root.findall(".//StopTime")[0].text),
        status=status,
        filename=filename,
    )


def read_recording_xml(recording_xml):
    # type: (Union[str,pathlib.Path]) -> RecordingMetadata
    """
        Read metadata from an Axis recording from e.g. a SD card or a NAS

        .. code-block:: python

            import vi3o

            metadata = vi3o.read_recording_xml("/path/to/recording.xml")
            print("Width: {}, Height: {}".format(metadata.width, metadata.height))
            print("Number of blocks: {}".format(len(metadata.blocks)))
    """
    if isinstance(recording_xml, str):
        recording_xml = compat.pathlib.Path(recording_xml)

    if not recording_xml.is_file():
        raise compat.FileNotFoundError("recording.xml must be a file")

    root = _read_xml(recording_xml)

    try:
        blocks = []
        for pth in recording_xml.parent.glob("*/*.xml"):
            blocks.append(_read_recording_block(pth))

    except (ValueError, xml.etree.ElementTree.ParseError) as error:
        raise RuntimeError("Failure parsing block {}: {}".format(pth, error))
    blocks.sort()

    return RecordingMetadata(
        recording_id=root.attrib["RecordingToken"],
        channel=int(root.findall(".//SourceToken")[0].text),
        start=_parse_axis_xml_timestamp(root.findall(".//StartTime")[0].text),
        stop=_parse_axis_xml_timestamp(root.findall(".//StopTime")[0].text),
        width=int(root.findall(".//Width")[0].text),
        height=int(root.findall(".//Height")[0].text),
        framerate=float(root.findall(".//Framerate")[0].text),
        blocks=tuple(blocks),
    )


class Recording(object):
    """
        Load video from a folder containing a Axis recording from e.g.
        a SD card or a NAS

        Axis stores videos in a blocked format together with XML metadata.
        The XML metadata contains time information which may be used to
        deduce the wall time of the recordings should the video streams
        themselves lack the proper metadata.

        Either use the convenience method `vi3o.Video`

        .. code-block:: python

            import vi3o
            recording = vi3o.Video("/path/to/recording.xml")

        Or load metadata manually using `read_recording_xml`

        .. code-block:: python

            import vi3o
            metadata = vi3o.read_recording_xml("/path/to/recording.xml")
            recording = Recording(metadata)


    """

    def __init__(self, metadata, **kwargs):
        if not metadata.blocks:
            raise RuntimeError("No recordings discovered!")

        # Preprocessing of all video blocks in folder - this may take some time...
        videos = []  # type: List[_VideoBlock]
        timestamp_offset = 0
        for blk in metadata.blocks:
            # Note: Video raises OSError if file was not found
            filename = str(blk.filename)
            videos.append(
                cat._VideoBlock(
                    video=filename,
                    timestamp_offset=timestamp_offset,
                    systime_offset=blk.start,
                )
            )
            # The first frame in the next block will start at timestamp offset
            timestamp_offset += blk.stop - blk.start

        self._video = cat.VideoCat(videos, **kwargs)

    def __len__(self):
        return len(self._video)

    @property
    def systimes(self):
        # type: () -> List[float]
        """ Return the systimes of all frames in recording """
        return self._video.systimes

    def __iter__(self):
        return iter(self._video)

    def __getitem__(self, item):
        return self._video[item]
